/* -*- coding: utf-8 -*-
 * vim: et:sw=4:sts=4
 */
/**
 * Copyright(c) 2017 Norwegian University of Science and Technology
 */
const
    log = require('@ntnu/logger')()('@ntnu/tia-neo4j'),
    neo = require('neo4j-driver').v1,
    resultGraph = require('./neo-query-result'),
    WriterMem = require('./neo-query-writer-mem'),
    MapperEmpty = require('./neo-query-mapper-empty'),
    util = require('util');

/* unindentCypher turns:
 *         MATCH
 *             (a)-->(b)
 *         RETURN
 *             a.foo AS foo,
 *             b.bar AS bar
 *
 * into:
 * MATCH
 *     (a)-->(b)
 * RETURN
 *     a.foo AS foo,
 *     b.bar AS bar
 */
const unindentCypher = (cypher) => {
    const indent = cypher.match(/ +(?=MATCH|MERGE|CREATE)|$/i)[0];
    return cypher.split('\n').map((x) => {
        return x.startsWith(indent) ? x.slice(indent.length) : x;
    }).join('\n').trim();
};

class Query {
    constructor(options, req) {
        this.options = Object.assign({
            mapper: new MapperEmpty(), //Mapper to map from js-on to js-on
            skip: 0,
            limit: 500,
            schema: {
                isMultiple: ['entityKeys', 'hasCar', 'hasName', 'role', 'hasContactinfo', 'roleQualifier', 'status', 'HAS_STILLING', 'registerkort', 'sourceSystem', 'masterNode', 'emneinfoType']
            },
            neo4j: {
                url: 'bolt://localhost',
                user: null,
                pass: null,
                retries: 1,
            },
            easeOf: 1,
        }, options);
        let auth;
        if (this.options.neo4j.user) {
            auth = neo.auth.basic(options.neo4j.user, options.neo4j.pass);
        }
        this.driver = neo.driver(this.options.neo4j.url, auth);

        if (this.options.limit < 0) {
            this.options.limit = 0;
        }

        if (this.options.skip < 0) {
            this.options.skip = 0;
        }

        if (!this.options.writer) {
            log.debug('No writer supplied, using internal writer!');
            this.data = {};
            this.options.writer = new WriterMem(this.data);
        }
        this.req = req;
        this.kill = false;
        req.on("close", () => { //Catch client in breaking connection
            this.kill = true;
        });
    }

    execute(cypher, params) {
        cypher = unindentCypher(cypher);
        params = params || {};
        log.debug({params}, cypher);

        const paginationPlaceholder = '<PAGINATION>'; //This will be replaced with skip and limit if placeholder is present in cypher-query
        const pagination = cypher.includes(paginationPlaceholder); //To avoid pagination when not applicable
        let eof = false;

        const runCypher = (cypher) => {
            let times = 0;
            const start = process.hrtime.bigint();
            const attemt = () => {
                const session = this.driver.session();
                return session.readTransaction((tx) => {
                    return tx.run(cypher, params);
                }).then((neoResult) => {
                    session.close();
                    log.debug(`queryTime: ${(process.hrtime.bigint() - start) / 1000000n} ms`);
                    return neoResult;
                }).catch((error) => {
                    session.close();
                    if (times >= this.options.neo4j.retries) {
                        log.warn(`Bailing out after ${times} retries! ${error}`);
                        throw error;
                    } else {
                        times++
                        log.warn(`Retry [${times}/${this.options.neo4j.retries}] on error! ${error}`);
                        return attemt();
                    }
                });
            };
            return attemt();
        };


        const paginate = (options) => {
            return new Promise((resolve) => {
                if (!pagination) {
                    resolve(cypher);
                    return;
                }
                let limit = 0,
                    skip = 0,
                    maxlimit = 1;

                if (options.limit && options.limit > 0) {
                    limit = options.limit;
                }
                if (options.skip && options.skip > 0) {
                    skip = options.skip;
                }
                if (options.maxlimit && options.maxlimit > 0) {
                    maxlimit = options.maxlimit
                }

                if (maxlimit < skip + limit) {
                    limit = maxlimit - skip;
                }

                const pageCypher = `${skip > 0 ? ' '+skip : ''}${limit > 0 ? ' '+limit : ''}`;
                if (pageCypher) {
                    log.debug(`skip: ${skip}, pagesize: ${limit}, maxlimit ${maxlimit}`);
                }
                resolve(cypher.replace(paginationPlaceholder, pageCypher));
            });
        };

        const increment = (query) => {
            return paginate(query.options)
            .then((cypher) => runCypher(cypher))
            .then((neoResult) => resultGraph(neoResult, query.options.schema))
            .then((queryResponse) => { //Query response from each increment
                if (queryResponse.length === 0) {
                    return;
                }
                //log.debug('Hits in result %s', queryResponse.length);
                //Calling the specified mapper, or using the default/empty one
                return this.options.mapper.map(queryResponse).then((mappedResult) => {
                    const limitNotReached = (!this.options.maxlimit || (query.options.skip + query.options.limit) < this.options.maxlimit); //When limits has not been reached

                    if (
                        !this.kill && //When client has not aborted
                        mappedResult.length > 0 //When there is data
                    ) {
                        //Writing to specified writer on each increment. If non specified, using the default memory storage writer.
                        return query.options.writer.write(mappedResult).then(() => {
                            if (limitNotReached && pagination) {
                                query.options.skip += query.options.limit;
                                //This might be a good idea if hammering the server isn't
                                /*const incr = () => {
                                    increment(query);
                                };
                                setTimeout(incr, query.options.easeOf);*/
                                return increment(query);
                            }
                        });
                    }
                });
            }).catch((error) => {
                if (!eof) { //to prevent multiple write/close on closed writer
                    return this.options.writer.write([`${error}`]).then(() => {
                        eof = true;
                        return this.options.writer.end(error);
                    });
                }
                log.error(`${error}`);
                throw error;
            });
        };
        return increment(this).then(() => {
            eof = true;
            return this.options.writer.end();
        }).catch((error) => {
            eof = true;
            return this.options.writer.end(error);
        });
    }
}

module.exports.Query = Query;
module.exports.WriterJSON = require('./neo-query-writer-json.js');
