const
    log = require('@ntnu/logger')()('neo-query-result'),
    neo = require('neo4j-driver').v1;

function getNeoProperty(prop) {
    return neo.isInt(prop) ? prop.toNumber() : prop;
}

const neoResult = (neoResult, schema) => { //Building js-on object from nodes and relations
    schema = schema || {};
    schema = {
        filter: schema.filter || [],
        fieldMap: schema.fieldMap || {},
        isMultiple: schema.isMultiple || [],
    };
    return new Promise((resolve) => {
        const relMap = new Map(); //Map of all relations, mapping to get rid of duplicates
        const endNodes = new Set(); //Tracking what nodes is related to, to be able to return those that isn't
        const nodeMap = new Map(); //Hashed up nodes

        const mapRelationship = (column) => {
            const relationship = {
                _identity: getNeoProperty(column.identity),
                _start: getNeoProperty(column.start),
                _end: getNeoProperty(column.end),
            };
            if (column.type) {
                relationship._type = column.type;
            }
            const relHash = `s${relationship._start}e${relationship._end}`; //Double relations with same start and end node, will be overwritten.
            relMap.set(relHash, relationship);
            endNodes.add(relationship._end);
        };
        const mapNode = (column) => {
            const node = {_identity: getNeoProperty(column.identity)};
            if (node._identity in nodeMap) {
                return;
            }

            if (typeof column.start !== 'undefined' && typeof column.end !== 'undefined') {
                node._start = getNeoProperty(column.start);
                node._end = getNeoProperty(column.end);
                if (column.type) {
                    node._type = column.type;
                }
            }
            for(let [key, value] of Object.entries(column.properties)) {
                node[key] = getNeoProperty(value);
            }
            nodeMap.set(node._identity, node);
        };
        const mapColumn = (column) => {
            const colType = Object.getPrototypeOf(column).constructor.name;
            if (Array.isArray(column)) {
                let mapThis = true;
                for (let _childColumn of column) {
                    if (!mapColumn(_childColumn)) {
                        mapThis = false;
                    }
                }
                if (!mapThis) {
                    return false;
                }
            } else if (colType === 'Node') {
                mapNode(column);
            } else if (colType === 'Relationship') {
                mapRelationship(column);
            } else if (colType === 'Path') {
                mapNode(column.start);
                mapNode(column.end);
                for (let segment of column.segments) {
                    mapNode(segment.start);
                    mapNode(segment.end);
                    mapRelationship(segment.relationship);
                }
            } else {
                return false;
            }
            return true;
        };

        //Hashing fragments
        let recCount = 0;
        for (let record of neoResult.records) {
            const rec = {};
            for (let i = record.length - 1; i >= 0; i--) {
                const key = record.keys[i];
                const column = record._fields[record._fieldLookup[key]];
                if (column && !mapColumn(column)) {
                    //No mapping, setting single value on column
                    rec[key] = getNeoProperty(column);
                }
            }
            if (Object.getOwnPropertyNames(rec).length > 0) {
                nodeMap.set(`r${recCount}`, rec);
            }
            recCount++;
        }

        //Stitching things together
        for (let rel of relMap.values()) {
            const node = nodeMap.get(rel._start);
            if (rel && node) {
                const endNode = nodeMap.get(rel._end);
                if (endNode) {
                    if (typeof node[rel._type] !== 'undefined') {
                        if (!Array.isArray(node[rel._type])) {
                            log.debug(`Overwriting [${rel._type}]`);
                            node[rel._type] = [node[rel._type]];
                        }
                        if (!node[rel_type].find((x) => x._identity === endNode._identity)) {
                            node[rel._type].push(endNode);
                        }
                    } else {
                        if (schema.isMultiple.includes(rel._type)) {
                            node[rel._type] = [endNode];
                        } else {
                            node[rel._type] = endNode;
                        }
                    }
                }
            }
        }

        const queryResponse = [];
        //Return what is not related
        for(let key of nodeMap.keys()) {
            if (!endNodes.has(key)) {
                queryResponse.push(nodeMap.get(key));
            }
        }

        //Cleanup
        for (let [nodeId, node] of nodeMap) {
            delete node._identity;
            delete node._checksum;
            delete node._id;
        }
        resolve(queryResponse);
    });
};
module.exports = neoResult;
